@@ -1,30 +1,33 @@
 import os
 import re
+import shlex
 import socket
 import time
+from datetime import timedelta
+from os import name as os_name
 from typing import Any, List, Optional, Union
 
 from robot.api import logger
 from robot.api.deco import keyword
 from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
-from robot.utils import Matcher
+from robot.utils import Matcher, secs_to_timestr, timestr_to_secs
 
 from .py3270 import Emulator
 
 
-class x3270(object):
+class X3270(object):
     def __init__(
         self,
         visible: bool,
-        timeout: int,
-        wait_time: float,
-        wait_time_after_write: float,
+        timeout: timedelta,
+        wait_time: timedelta,
+        wait_time_after_write: timedelta,
         img_folder: str,
     ) -> None:
         self.visible = visible
-        self.timeout = timeout
-        self.wait = wait_time
-        self.wait_write = wait_time_after_write
+        self.timeout = self._convert_timeout(timeout)
+        self.wait = self._convert_timeout(wait_time)
+        self.wait_write = self._convert_timeout(wait_time_after_write)
         self.imgfolder = img_folder
         self.mf: Emulator = None  # type: ignore
         # Try Catch to run in Pycharm, and make a documentation in libdoc with no error
@@ -33,15 +36,20 @@ class x3270(object):
         except RobotNotRunningError:
             self.output_folder = os.getcwd()
 
+    def _convert_timeout(self, time):
+        if isinstance(time, timedelta):
+            return time.total_seconds()
+        return timestr_to_secs(time, round_to=None)
+
     @keyword("Change Timeout")
-    def change_timeout(self, seconds: int) -> None:
+    def change_timeout(self, seconds: timedelta) -> None:
         """
         Change the timeout for connection in seconds.
 
         Example:
-            | Change Timeout | 3 |
+            | Change Timeout | 3 seconds |
         """
-        self.timeout = seconds
+        self.timeout = self._convert_timeout(seconds)
 
     @keyword("Open Connection")
     def open_connection(
@@ -51,41 +59,150 @@ class x3270(object):
         port: int = 23,
         extra_args: Optional[Union[List[str], os.PathLike]] = None,
     ):
-        """Create a connection to IBM3270 mainframe with the default port 23. To make a connection with the mainframe
-        you only must inform the Host. You can pass the Logical Unit Name and the Port as optional.
+        """Create a connection to an IBM3270 mainframe with the default port 23.
+        To establish a connection, only the hostname is required. Optional parameters include logical unit name (LU) and port.
+
+        Additional configuration data can be provided through the `extra_args` parameter.
+        `extra_args` accepts either a list or a path to a file containing [https://x3270.miraheze.org/wiki/Category:Command-line_options|x3270 command line options].
 
-        If you wish, you can provide further configuration data to ``extra_args``. ``extra_args`` takes in a list,
-        or a path to a file, containing [https://x3270.miraheze.org/wiki/Category:Command-line_options|x3270 command line options].
         Entries in the argfile can be on one line or multiple lines. Lines starting with "#" are considered comments.
+        Arguments containing whitespace must be enclosed in single or double quotes.
+
         | # example_argfile_oneline.txt
-        | --accepthostname myhost.com
+        | -accepthostname myhost.com
 
         | # example_argfile_multiline.txt
-        | --accepthostname myhost.com
+        | -xrm "wc3270.acceptHostname: myhost.com"
         | # this is a comment
-        | --charset french
+        | -charset french
+        | -port 992
 
-        Please make sure the arguments you are providing are available for your specific x3270 application and version.
+        Please ensure that the arguments provided are available for your specific x3270 application and version.
+        Refer to the [https://x3270.miraheze.org/wiki/Wc3270/Command-line_options|wc3270 command line options] for a subset of available options.
+
+        Note: If you specify the port with the `-port` command-line option in `extra_args` (or use the -xrm resource command for it),
+        it will take precedence over the `port` argument provided in the `Open Connection` keyword.
 
         Example:
             | Open Connection | Hostname |
             | Open Connection | Hostname | LU=LUname |
             | Open Connection | Hostname | port=992 |
-            | ${extra_args}   | Create List | --accepthostname | myhost.com | --cafile | ${CURDIR}/cafile.crt |
+            | @{extra_args}   | Create List | -accepthostname | myhost.com | -cafile | ${CURDIR}/cafile.crt |
+            | Append To List  | ${extra_args} | -port | 992 |
             | Open Connection | Hostname | extra_args=${extra_args} |
             | Open Connection | Hostname | extra_args=${CURDIR}/argfile.txt |
         """
-        self.host = host
-        self.lu = LU
-        self.port = port
-        if self.lu:
-            self.credential = "%s@%s:%s" % (self.lu, self.host, self.port)
-        else:
-            self.credential = "%s:%s" % (self.host, self.port)
         if self.mf:
             self.close_connection()
+        extra_args = self._process_args(extra_args)
         self.mf = Emulator(self.visible, self.timeout, extra_args)
-        self.mf.connect(self.credential)
+        host_string = f"{LU}@{host}" if LU else host
+        if self._port_in_extra_args(extra_args):
+            if port != 23:
+                logger.warn(
+                    "The connection port has been specified both in the `port` argument and in `extra_args`. "
+                    "The port specified in `extra_args` will take precedence over the `port` argument. "
+                    "To avoid this warning, you can either remove the port command-line option from `extra_args`, "
+                    "or leave the `port` argument at its default value of 23."
+                )
+            self.mf.connect(host_string)
+        else:
+            self.mf.connect(f"{host_string}:{port}")
+
+    def _process_args(self, args) -> list:
+        processed_args = []
+        if not args:
+            return []
+        elif isinstance(args, list):
+            processed_args = args
+        elif isinstance(args, os.PathLike) or isinstance(args, str):
+            with open(args) as file:
+                for line in file:
+                    if line.lstrip().startswith("#"):
+                        continue
+                    for arg in shlex.split(line):
+                        processed_args.append(arg)
+        return processed_args
+
+    def _port_in_extra_args(self, args) -> bool:
+        if not args:
+            return False
+        for arg in args:
+            if arg == "-port" or ".port" in arg:
+                return True
+        return False
+
+    @keyword("Open Connection From Session File")
+    def open_connection_from_session_file(self, session_file: os.PathLike):
+        """Create a connection to an IBM3270 mainframe using a [https://x3270.miraheze.org/wiki/Session_file|session file].
+
+        The session file contains [https://x3270.miraheze.org/wiki/Category:Resources|resources (settings)] for a specific host session.
+        The only mandatory setting required to establish the connection is the [https://x3270.miraheze.org/wiki/Hostname_resource|hostname resource].
+
+        This keyword is an alternative to `Open Connection`. Please note that the Robot-Framework-Mainframe-3270-Library
+        currently only supports model "2". Specifying any other model will result in a failure.
+
+        For more information on session file syntax and detailed examples, please consult the [https://x3270.miraheze.org/wiki/Session_file|x3270 wiki].
+
+        Example:
+        | Open Connection From Session File | ${CURDIR}/session.wc3270 |
+
+        where the content of `session.wc3270` is:
+        | wc3270.hostname: myhost.com:23
+        | wc3270.model: 2
+        """
+        if self.mf:
+            self.close_connection()
+        self._check_session_file_extension(session_file)
+        self._check_contains_hostname(session_file)
+        self._check_model(session_file)
+        if os_name == "nt" and self.visible:
+            self.mf = Emulator(self.visible, self.timeout)
+            self.mf.connect(str(session_file))
+        else:
+            self.mf = Emulator(self.visible, self.timeout, [str(session_file)])
+
+    def _check_session_file_extension(self, session_file):
+        file_extension = str(session_file).rsplit(".")[-1]
+        expected_extensions = {
+            ("nt", True): "wc3270",
+            ("nt", False): "ws3270",
+            ("posix", True): "x3270",
+            ("posix", False): "s3270",
+        }
+        expected_extension = expected_extensions.get((os_name, self.visible))
+        if file_extension != expected_extension:
+            raise ValueError(
+                f"Based on the emulator that you are using, "
+                f'the session file extension has to be ".{expected_extension}", '
+                f'but it was ".{file_extension}"'
+            )
+
+    def _check_contains_hostname(self, session_file):
+        with open(session_file) as file:
+            if "hostname:" not in file.read():
+                raise ValueError(
+                    "Your session file needs to specify the hostname resource "
+                    "to set up the connection. "
+                    "An example for wc3270 looks like this: \n"
+                    "wc3270.hostname: myhost.com\n"
+                )
+
+    def _check_model(self, session_file):
+        with open(session_file) as file:
+            pattern = re.compile(r"[wcxs3270.*]+model:\s*([327892345E-]+)")
+            match = pattern.findall(file.read())
+            if not match:
+                return
+            elif match[-1] == "2":
+                return
+            else:
+                raise ValueError(
+                    f'Robot-Framework-Mainframe-3270-Library currently only supports model "2", '
+                    f'the model you specified in your session file was "{match[-1]}". '
+                    f'Please change it to "2", using either the session wizard if you are on Windows, '
+                    f'or by editing the model resource like this "*model: 2"'
+                )
 
     @keyword("Close Connection")
     def close_connection(self) -> None:
@@ -97,7 +214,7 @@ class x3270(object):
         self.mf = None  # type: ignore
 
     @keyword("Change Wait Time")
-    def change_wait_time(self, wait_time: float) -> None:
+    def change_wait_time(self, wait_time: timedelta) -> None:
         """To give time for the mainframe screen to be "drawn" and receive the next commands, a "wait time" has been
         created, which by default is set to 0.5 seconds. This is a sleep applied AFTER the following keywords:
 
@@ -110,13 +227,14 @@ class x3270(object):
         If you want to change this value, just use this keyword passing the time in seconds.
 
         Example:
-            | Change Wait Time | 0.1 |
-            | Change Wait Time | 2 |
+            | Change Wait Time | 0.5              |
+            | Change Wait Time | 200 milliseconds |
+            | Change Wait Time | 0:00:01.500      |
         """
-        self.wait = wait_time
+        self.wait = self._convert_timeout(wait_time)
 
     @keyword("Change Wait Time After Write")
-    def change_wait_time_after_write(self, wait_time_after_write: float) -> None:
+    def change_wait_time_after_write(self, wait_time_after_write: timedelta) -> None:
         """To give the user time to see what is happening inside the mainframe, a "wait time after write" has
         been created, which by default is set to 0 seconds. This is a sleep applied AFTER sending a string in these
         keywords:
@@ -131,10 +249,11 @@ class x3270(object):
         Note: This keyword is useful for debug purpose
 
         Example:
-            | Change Wait Time After Write | 0.5 |
-            | Change Wait Time After Write | 2 |
+            | Change Wait Time After Write | 1             |
+            | Change Wait Time After Write | 0.5 seconds   |
+            | Change Wait Time After Write | 0:00:02       |
         """
-        self.wait_write = wait_time_after_write
+        self.wait_write = self._convert_timeout(wait_time_after_write)
 
     @keyword("Read")
     def read(self, ypos: int, xpos: int, length: int) -> str:
@@ -201,18 +320,25 @@ class x3270(object):
             logger.warn('Screenshots will be saved in "%s"' % self.imgfolder)
 
     @keyword("Take Screenshot")
-    def take_screenshot(self, height: int = 410, width: int = 670) -> None:
+    def take_screenshot(
+        self, height: int = 410, width: int = 670, filename_prefix: str = "screenshot"
+    ) -> str:
         """Generate a screenshot of the IBM 3270 Mainframe in a html format. The
         default folder is the log folder of RobotFramework, if you want change see the `Set Screenshot Folder`.
 
         The Screenshot is printed in a iframe log, with the values of height=410 and width=670, you
-        can change this values passing them to the keyword.
+        can change these values by passing them to the keyword.
+
+        The file name prefix can be set, the default is "screenshot".
+
+        The full file path is returned.
 
         Example:
-            | Take Screenshot |
+            | ${filepath} | Take Screenshot |
+            | ${filepath} | Take Screenshot | height=500 | width=700 |
             | Take Screenshot | height=500 | width=700 |
+            | Take Screenshot | filename_prefix=MyScreenshot |
         """
-        filename_prefix = "screenshot"
         extension = "html"
         filename_sufix = round(time.time() * 1000)
         filepath = os.path.join(
@@ -225,6 +351,7 @@ class x3270(object):
             level="INFO",
             html=True,
         )
+        return filepath
 
     @keyword("Wait Field Detected")
     def wait_field_detected(self) -> None:
@@ -301,7 +428,7 @@ class x3270(object):
         Example:
                | Send PF | 3 |
         """
-        self.mf.exec_command(("PF(" + PF + ")").encode("utf-8"))
+        self.mf.exec_command(("PF({0})").format(PF).encode("utf-8"))
         time.sleep(self.wait)
 
     @keyword("Write")
@@ -365,22 +492,25 @@ class x3270(object):
             time.sleep(self.wait)
 
     @keyword("Wait Until String")
-    def wait_until_string(self, txt: str, timeout: int = 5) -> str:
+    def wait_until_string(
+        self, txt: str, timeout: timedelta = timedelta(seconds=5)
+    ) -> str:
         """Wait until a string exists on the mainframe screen to perform the next step. If the string does not appear in
         5 seconds, the keyword will raise an exception. You can define a different timeout.
 
         Example:
             | Wait Until String | something |
-            | Wait Until String | something | timeout=10 |
+            | Wait Until String | something | 10 |
+            | Wait Until String | something | 15 s |
+            | Wait Until String | something | 0:00:15 |
         """
-        max_time = time.ctime(int(time.time()) + timeout)
-        while time.ctime(int(time.time())) < max_time:
+        timeout = self._convert_timeout(timeout)
+        max_time = time.time() + timeout  # type: ignore
+        while time.time() < max_time:
             result = self._search_string(str(txt))
             if result:
                 return txt
-        raise Exception(
-            'String "' + txt + '" not found in ' + str(timeout) + " seconds"
-        )
+        raise Exception(f'String "{txt}" not found in {secs_to_timestr(timeout)}')
 
     def _search_string(self, string: str, ignore_case: bool = False) -> bool:
         """Search if a string exists on the mainframe screen and return True or False."""
@@ -409,10 +539,10 @@ class x3270(object):
 
         Example:
             | Page Should Contain String | something |
-            | Page Should Contain String | someTHING | ignore_case=True |
+            | Page Should Contain String | someTHING | ignore_case=True                |
             | Page Should Contain String | something | error_message=New error message |
         """
-        message = 'The string "' + txt + '" was not found'
+        message = f'The string "{txt}" was not found'
         if error_message:
             message = error_message
         if ignore_case:
@@ -420,7 +550,7 @@ class x3270(object):
         result = self._search_string(txt, ignore_case)
         if not result:
             raise Exception(message)
-        logger.info('The string "' + txt + '" was found')
+        logger.info(f'The string "{txt}" was found')
 
     @keyword("Page Should Not Contain String")
     def page_should_not_contain_string(
@@ -437,7 +567,7 @@ class x3270(object):
             | Page Should Not Contain String | someTHING | ignore_case=True |
             | Page Should Not Contain String | something | error_message=New error message |
         """
-        message = 'The string "' + txt + '" was found'
+        message = f'The string "{txt}" was found'
         if error_message:
             message = error_message
         if ignore_case:
@@ -464,7 +594,7 @@ class x3270(object):
             | Page Should Contain Any String | ${list_of_string} | ignore_case=True |
             | Page Should Contain Any String | ${list_of_string} | error_message=New error message |
         """
-        message = 'The strings "' + str(list_string) + '" were not found'
+        message = f'The strings "{list_string}" were not found'
         if error_message:
             message = error_message
         if ignore_case:
@@ -547,7 +677,7 @@ class x3270(object):
             result = self._search_string(string, ignore_case)
             if result:
                 if message is None:
-                    message = 'The string "' + string + '" was found'
+                    message = f'The string "{string}" was found'
                 raise Exception(message)
 
     @keyword("Page Should Contain String X Times")
@@ -580,7 +710,7 @@ class x3270(object):
             if message is None:
                 message = f'The string "{txt}" was not found "{number}" times, it appears "{number_of_times}" times'
             raise Exception(message)
-        logger.info('The string "' + txt + '" was found "' + str(number) + '" times')
+        logger.info(f'The string "{txt}" was found "{number}" times')
 
     @keyword("Page Should Match Regex")
     def page_should_match_regex(self, regex_pattern: str) -> None:
@@ -594,7 +724,7 @@ class x3270(object):
         """
         page_text = self._read_all_screen()
         if not re.findall(regex_pattern, page_text, re.MULTILINE):
-            raise Exception('No matches found for "' + regex_pattern + '" pattern')
+            raise Exception(f'No matches found for "{regex_pattern}" pattern')
 
     @keyword("Page Should Not Match Regex")
     def page_should_not_match_regex(self, regex_pattern: str) -> None:
@@ -608,9 +738,7 @@ class x3270(object):
         """
         page_text = self._read_all_screen()
         if re.findall(regex_pattern, page_text, re.MULTILINE):
-            raise Exception(
-                'There are matches found for "' + regex_pattern + '" pattern'
-            )
+            raise Exception(f'There are matches found for "{regex_pattern}" pattern')
 
     @keyword("Page Should Contain Match")
     def page_should_contain_match(
@@ -644,7 +772,7 @@ class x3270(object):
         result = matcher.match(all_screen)
         if not result:
             if message is None:
-                message = 'No matches found for "' + txt + '" pattern'
+                message = f'No matches found for "{txt}" pattern'
             raise Exception(message)
 
     @keyword("Page Should Not Contain Match")
@@ -679,7 +807,7 @@ class x3270(object):
         result = matcher.match(all_screen)
         if result:
             if message is None:
-                message = 'There are matches found for "' + txt + '" pattern'
+                message = f'There are matches found for "{txt}" pattern'
             raise Exception(message)
 
     def _read_all_screen(self) -> str:
@@ -702,11 +830,11 @@ class x3270(object):
             result = self._search_string(string, ignore_case)
             if not should_match and result:
                 if message is None:
-                    message = 'The string "' + string + '" was found'
+                    message = f'The string "{string}" was found'
                 raise Exception(message)
             elif should_match and not result:
                 if message is None:
-                    message = 'The string "' + string + '" was not found'
+                    message = f'The string "{string}" was not found'
                 raise Exception(message)
 
     @staticmethod
